local addonName, addonTable = ...
-- Debugging to tekDebug
local debug = false
--[===[@debug@
debug = true
--@end-debug@]===]

local name, _, _, enabled = GetAddOnInfo("tekDebug")
local player = UnitName("player")
local enabled = GetAddOnEnableState(player, "tekDebug") == 2
if enabled and not IsAddOnLoaded("tekDebug") then
  local succ, err = LoadAddOn(name)
end
local debugf = debug and tekDebug and tekDebug:GetFrame(addonName)
local function DebugPrint(...) if debugf then debugf:AddMessage(string.join(", ", tostringall(...))) end end

local strfind, strformat, strgsub, strlower, strmatch, tostring = string.find, string.format, string.gsub, string.lower, strmatch, tostring
local _, date, ipairs, next, pairs, select, time, tinsert, tsort, type, wipe = _, date, ipairs, next, pairs, select, time, table.insert, table.sort, type, wipe

-- LibDataBroker
local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local LDBIcon = ldb and LibStub("LibDBIcon-1.0",true)
local LibQTip = LibStub('LibQTip-1.0')
local BF = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceConsole-3.0", "AceEvent-3.0")

if debug then
  Broker_Followers = BF
end
local L = LibStub("AceLocale-3.0"):GetLocale(addonName, false)
local DB_VERSION = 1
BF.followerList = {}
BF.collectedFollowersList = {}
BF.reverseCounterLookup = {}
BF.formatStrings = {
  rarityColors = {},
}
BF.threatStrings = {
  [1] = "Wild Aggression",
  [2] = "Massive Strike",
  [3] = "Group Damage",
  [4] = "Magic Debuff",
  [6] = "Danger Zones",
  [7] = "Minion Swarms",
  [8] = "Powerful Spell",
  [9] = "Deadly Minions",
  [10] = "Timed Battle",
  [11] = "Orc",
  [12] = "Ogre",
  [13] = "Undead",
  [14] = "Arakkoa",
  [15] = "Beast",
  [16] = "Demon",
  [17] = "Breaker",
  [18] = "Fury",
  [19] = "Aberration",
  [20] = "Primal",
  [21] = "Mountains",
  [22] = "Desert",
  [23] = "Snow",
  [24] = "Underground",
  [25] = "Jungle",
  [26] = "Forest",
  [27] = "Town",
  [28] = "Swamp",
  [29] = "Plains",
}
--[[====
Threats
6:  Danger Zones      Environments            Races
9:  Deadly Minions    22: Desert              19: Aberration
3:  Group Damage      26: Forest              14: Arakkoa
4:  Magic Debuff      58: Hearthstone Table   15: Beast
2:  Massive Strike    25: Jungle              17: Breaker
7:  Minion Swarms     21: Mountains           16: Demon
8:  Powerful Spell    29: Plains              18: Fury
10: Timed Battle      23: Snow                12: Ogre
1:  Wild Aggression   28: Swamp               11: Orc
                      27: Town                20: Primal
                      24: Underground         13: Undead
====]]--

BF.specs = {
  [2] = {
    --Blood Death Knight
    1,
    2,
    7,
    8,
    10,
  },
  ["Frost Death Knight"] = {
    --Frost Death Knight
    1,
    4,
    7,
    8,
    10,
  },
  [4] = {
    --Unholy Death Knight
    1,
    2,
    7,
    8,
    10,
  },
  [5] = {
    --Balance Druid
    6,
    7,
    9,
    10,
  },
  [7] = {
    --Feral Druid
    1,
    2,
    6,
    10,
  },
  ["Guardian Druid"] = {
    --Guardian Druid
    1,
    2,
    6,
    9,
  },
  [9] = {
    --Restoration Druid
    3,
    4,
    7,
    9,
  },
  [10] = {
    --Beast Master Hunter
    1,
    6,
    7,
    9,
    10,
  },
  [12] = {
    --Marksmanship Hunter
    6,
    7,
    8,
    9,
    10,
  },
  [13] = {
    --Survival Hunter
    2,
    6,
    7,
    9,
    10,
  },
  [14] = {
    --Arcane Mage
    6,
    8,
    9,
    10,
  },
  [15] = {
    --Fire Mage
    6,
    7,
    8,
    9,
    10,
  },
  ["Frost Mage"] = {
    --Frost Mage
    2,
    7,
    8,
    9,
    10,
  },
  [17] = {
    --Brewmaster Monk
    1,
    2,
    3,
    6,
    9,
  },
  [18] = {
    --Mistweaver Monk
    3,
    4,
    6,
    8,
  },
  [19] = {
    --Windwalker Monk
    1,
    6,
    8,
    9,
    10,
  },
  [20] = {
    --Holy Paladin
    3,
    4,
    8,
    9,
  },
  ["Protection Paladin"] = {
    --Protection Paladin
    1,
    2,
    4,
    8,
    9,
  },
  [22] = {
    --Retribution Paladin
    2,
    7,
    8,
    9,
    10,
  },
  ["Discipline Priest"] = {
    --Discipline Priest
    3,
    4,
    6,
    9,
  },
  [24] = {
    --Holy Priest
    3,
    4,
    6,
    7,
  },
  [25] = {
    --Shadow Priest
    4,
    6,
    7,
    9,
    10,
  },
  [26] = {
    --Assassination Rogue
    2,
    6,
    8,
    9,
    10,
  },
  [27] = {
    --Combat Rogue
    6,
    7,
    8,
    9,
    10,
  },
  [28] = {
    --Subtlety Rogue
    2,
    6,
    7,
    8,
    9,

  },
  [29] = {
    --Elemental Shaman
    3,
    7,
    8,
    9,
    10,
  },
  [30] = {
    --Enhancement Shaman
    3,
    6,
    7,
    9,
    10,
  },
  [31] = {
    --Restoration Shaman
    3,
    4,
    7,
    8,
  },
  ["Affliction Warlock"] = {
    --Affliction Warlock
    4,
    7,
    8,
    9,
    10,
  },
  [33] = {
    --Demonology Warlock
    2,
    7,
    8,
    10,
  },
  [34] = {
    --Destruction Warlock
    3,
    8,
    9,
    10,
  },
  [35] = {
    --Arms Warrior
    1,
    6,
    7,
    8,
    10,
  },
  [37] = {
    --Fury Warrior
    2,
    6,
    7,
    8,
    10,
  },
  [38] = {
    --Protection Warrior
    1,
    2,
    6,
    7,
    8,
  },
}
local defaults = {
  profile = {
    version = 1,
    availableCountersToolstip = true,
    showRerollPercent         = true,
    iconSize                  = 16,
    clickToShow               = 2,
    colorMatrix               = true,
    showInactivesMatrix       = true,
    minimapIcon = {
      hide = true,
      minimapPos = 220,
      radius = 80,
    },
  },
  global = {
    version = 1,
  },
}

function BF:OnInitialize() -- ADDON_LOADED, only saved variables are guaranteed to be present
  self.db = LibStub("AceDB-3.0"):New("Broker_Follower_DB", defaults, true)
  self:MigrateSavedData()

  LibStub("AceConfig-3.0"):RegisterOptionsTable(addonName, self.GenerateOptions)
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(addonName)

  self.BFLDB = ldb:NewDataObject(addonName, {
    type = "data source",
    text = "",
    icon = "Interface\\Icons\\inv_misc_lockboxghostiron",
    OnClick = function(clickedframe, button)
      if button == "RightButton" then
        InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
      else
        -- 1 = click2show
        -- 2 = hover2show
        if self.db.profile.clickToShow == 2 then
          self.db.profile.showInactivesMatrix = not self.db.profile.showInactivesMatrix
          self:PrepareTooltip(clickedframe, not IsShiftKeyDown())
        else
          -- Click 2 Show handling
          local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
          if tooltip:IsShown() then
            -- Tooltip is shown, hide it on click
            LibQTip:Release(BF.tooltip)
            BF.tooltip.Columns = nil
            BF.tooltip.Rows = nil
            self.tooltip = nil
          else
            -- No tooltip, show it on click
            self:PrepareTooltip(clickedframe, not IsShiftKeyDown())
          end
        end
      end
    end,
    OnEnter = function(tt)
      -- 1 = click2show
      -- 2 = hover2show
      if self.db.profile.clickToShow == 1 then return end
      self:PrepareTooltip(tt, not IsShiftKeyDown())
    end,
    OnLeave = function(tt)
      -- 1 = click2show
      -- 2 = hover2show
      if self.db.profile.clickToShow == 1 then return end
      LibQTip:Release(BF.tooltip)
      BF.tooltip.Columns = nil
      BF.tooltip.Rows = nil
      self.tooltip = nil
    end,
  })

  LDBIcon:Register(addonName, self.BFLDB, self.db.profile.minimapIcon)
  self:UpdateFollowerCounters()
  
  
  -- Update the .format strings 
  self.formatStrings.counterIconTolltip = "|T%s:"..self.db.profile.iconSize..":"..self.db.profile.iconSize..":0:0:64:64:5:59:5:59|t"
  for i=0,7 do
    self.formatStrings.rarityColors[i] = "|c"..select(4,GetItemQualityColor(i)).."%s|r"
  end
end

-- Options Handling

function BF:GenerateOptions()
  if not BF.options then

    local function newCounter()
      local i = 0
      return function( inc )
          i = i + 1
          return i
        end
    end
    local count = newCounter()

    BF.options = {
      name = addonName,
      type = 'group',
      handler = BF,
      args = {
        genconfig = {
          name = L["General Settings"],
          type = 'group',
          order = 1,
          get = function(info) return BF.db.profile[info[#info]] end,
          set = function(info, value) BF.db.profile[info[#info]] = value end,
          args = {
            showRerollPercent = {
              type = "toggle",
              name = L["Reroll column"],
              desc = L["In the list tooltip add a column at the front showing the % chance of getting a not-yet-acquired threat counter combination for epic followers when using a reroll certificate"],
              order = count(),
            },
            iconSize = {
              type = "range",
              min = 4,
              max = 128,
              step = 1,
              name = L["Tooltip: Icon size"],
              order = count(),
              set = function(info, value) 
                BF.db.profile[info[#info]] = value 
                BF.formatStrings.counterIconTolltip = "|T%s:"..value..":"..value..":0:0:64:64:5:59:5:59|t"
                BF:RedrawTooltip()
              end,
              
            },
            clickToShow = {
              type = "select",
              style = "dropdown",
              values = { 
                [1] = L["Click"],
                [2] = L["Hover"],
              },
              name = L["Show tooltip on:"],
              desc = L["Either make the tooltip pop up when you hover the icon or make it a toggle by clicking the icon"],
              order = count(),
            },
            colorMatrix = {
              type = "toggle",
              name = L["Color followers"],
              desc = L["Color followers in the matrix view by itemlevel. (615+: Green, 630+: Blue, 645+: Epic)"],
              order = count(),
            },
            showInactivesMatrix = {
              type = "toggle",
              name = L["Count Inactives"],
              desc = L["Also count inactive followers in matrix view"],
              order = count(),
            },
            availableCountersToolstip = {
              type = "toggle",
              name = L["Follower Hovertooltip"],
              desc = L["When hovering over the followerporait in the follower tab or while recruiting in the inn then display small tooltip with all available counters for this follower"],
              order = count(),
            },
            minimapIcon = {
              type = "toggle",
              name = L["Show minimap button"],
              order = count(),
              get = function(info) return not BF.db.profile.minimapIcon.hide end,
              set = function(info, value)
                BF.db.profile.minimapIcon.hide = not value
                if value then LDBIcon:Show(addonName) else LDBIcon:Hide(addonName) end
              end,
            },
          },
        },
      },
    }
  end
  return BF.options
end


local mechanics = {}
local q = C_Garrison.GetFollowerAbilityCounterMechanicInfo
for _, aid in pairs({11, 100, 168, 148, 160, 101, 105, 157, 122}) do
  local mid, _, tex = q(aid)
  mechanics[mid], mechanics[tex:lower():gsub("%.blp$","")] = aid, aid
end

local function getMechanicInfo(mid)
  if populateMechanics then
    populateMechanics = populateMechanics()
  end
  if mechanics[mid] then
    return C_Garrison.GetFollowerAbilityCounterMechanicInfo(mechanics[mid])
  end
end



function BF:PrepareTooltip(anchorFrame, tableInsteadOfList)
  local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
  self.tooltip = tooltip
  self:DrawTooltip(tooltip, tableInsteadOfList)
  tooltip:SmartAnchorTo(anchorFrame)
  tooltip:Show()
end

function BF:RedrawTooltip()
  -- Check if we actually view matrix tooltip

  local tooltip = LibQTip:Acquire("Broker_Followers_Tooltip", 1, "LEFT")
  -- Only redraw if the tooltip is shown
  if not tooltip:IsShown() then return end

  self.tooltip = tooltip
  self:DrawTooltip(tooltip, self.matrixIsUsed)
end

function BF:DrawTooltip(tooltip, tableInsteadOfList)
  UIParentLoadAddOn("Blizzard_GarrisonUI")

  self:UpdateFollowerCounters()
  tooltip:Clear()
  tooltip.Rows    = {}
  local function sortFollowers(f1, f2)
    local f1a, f2a = BF.collectedFollowersList[f1.followerID], BF.collectedFollowersList[f2.followerID]
    if f1.isCollected and f2.isCollected  then
      for _,i in pairs({1,2,3,4,6,7,8,9,10}) do
      --for i=1,10 do
        if f1a[i] and not f2a[i] then
          return true
        end
        if not f1a[i] and f2a[i] then
          return false
        end
      end
      return f1.name < f2.name
    else
      if not f1.isCollected and not f2.isCollected then
        return f1.name > f2.name
      else
        return f1.isCollected
      end
    end
  end

  local function sortAbilities(a1, a2)
    if not a1.isTrait and a2.isTrait  then
      return true
    end
    if a1.isTrait and  not a2.isTrait  then
      return false
    end
    local a1c = next(a1.counters)
    local a2c = next(a2.counters)
    if a1.isTrait and a2.isTrait  then
      if a1c or a2c then
        return (a1c or 0) > (a2c or 0)
      else
        return a1.name > a2.name
      end
    end
    a1c = next(a1.counters)
    a2c = next(a2.counters)
    return a1c < a2c
  end
  if not tooltip.Columns then
    tooltip.Columns = {}
  end
  if not tooltip.Rows then
    tooltip.Rows = {}
  end
  local function MyAddColumn(nameType, displayName)
    if nameType then
      if not displayName then
        BF:Debug("MyAddColumn: missing displayName for", nameType)
        displayName = nameType
      end
      if not formatFunc then
        --formatFunc = CellFormat.Default
      end

      if not tooltip.Columns[nameType] then
        local colindex = tooltip:AddColumn("CENTER")
        tooltip:SetCell(1, colindex, displayName)

        tooltip.Columns[nameType] = { col = colindex }
      else
        BF:Debug("a column already exists for", nameType)
      end
    end
  end
  local function AddSpanner(name, abbr, first, last)
    -- TODO - can we, from the size of span and maybe the character count of
    -- name make a guess as to whether we could use the name instead of Abbr?
    if first <= last then
      tooltip:SetCell(1, first, abbr, nil, nil, 1 + last - first)
    end
  end
  colors = {
    green = {r=0, g=1, b=0, a=1},
    white = {r=1, g=1, b=1, a=1},
    red = {r=1, g=0, b=0, a=1},

    lightGray = {r=(108/255), g=(115/255), b=(120/255), a=1},
    green = {r=(32/255), g=(255/255), b=(32/255), a=1},
    red = {r=(255/255), g=(25/255), b=(25/255), a=1},
    yellow = {r=(255/255), g=(228/255), b=(0/255), a=1},

    --lightGray = {r=0.25, g=0.25, b=0.25, a=1},
    darkGray = {r=0.1, g=0.1, b=0.1, a=1},
    lineGrey = {r=(108/255), g=(115/255), b=(120/255), a=1},
    --lightGray = {r=(82/255), g=(91/255), b=(97/255), a=1},
  }
  local function AddSeparator(tooltip)
    tooltip:AddSeparator(1, colors.lineGrey.r, colors.lineGrey.g, colors.lineGrey.b, colors.lineGrey.a)
  end

  local globaldb = self.db.global
  local profiledb = self.db.profile

  if tableInsteadOfList then
    self.matrixIsUsed = true
    -- 9x9 table, each cell showing the number of follower who counters both
    -- threats from column and row

    if tooltip:GetLineCount() < 3 then
      tooltip.Rows["icon"] = tooltip.Rows["icon"] or tooltip:AddHeader()
      tooltip.Rows["sum"]  = tooltip.Rows["sum"]  or tooltip:AddHeader()
    end
    MyAddColumn("icon","")
    tooltip:SetCell( 2,1,"")

    -- Setup the tooltip: Add the textures to the cells of the irst row and
    -- column, then "" all the other cells so we start with a clean matrix
    for k=1,10 do
      local _, _, tex = getMechanicInfo(k)
      if tex then
        tex = string.format(self.formatStrings.counterIconTolltip, tex)
        local row = tooltip.Rows[k] or tooltip:AddHeader()
        --AddSeparator(tooltip)
        MyAddColumn(tostring(k))
        tooltip:SetCell( 1, tooltip.Columns[tostring(k)].col, tex)
        tooltip:SetCell( row, 1, tex)
        tooltip.Rows[k] = row
        for j=1,10 do
          if tooltip.Columns[tostring(j)]then
            tooltip:SetCell( tooltip.Rows[k], tooltip.Columns[tostring(j)].col, "")
          end
        end
      end
    end
    local counterSum = {}
    tooltip.counterDualSum = {}
    local function addDualCounter( c1, c2, iLevel)
      local quality = 1
      if iLevel >= 615 then
        quality = 2
      end
      if iLevel >= 630 then
        quality = 3
      end
      if iLevel >= 645 then
        quality = 4
      end
      BF:Debug( c1, c2, iLevel, quality)
      tooltip.counterDualSum[c1]     = tooltip.counterDualSum[c1] or {}
      tooltip.counterDualSum[c2]     = tooltip.counterDualSum[c2] or {}
      tooltip.counterDualSum[c2][c1] = tooltip.counterDualSum[c2][c1] or {}
      tooltip.counterDualSum[c1][c2] = tooltip.counterDualSum[c1][c2] or {}
      tooltip.counterDualSum[c2][c1].num     = ( tooltip.counterDualSum[c2][c1].num or 0 ) + 1
      tooltip.counterDualSum[c1][c2].num     = ( tooltip.counterDualSum[c1][c2].num or 0 ) + 1
      tooltip.counterDualSum[c2][c1].quality = math.max( tooltip.counterDualSum[c2][c1].quality or 0,  quality )
      tooltip.counterDualSum[c1][c2].quality = math.max( tooltip.counterDualSum[c1][c2].quality or 0,  quality )
    end

    table.sort(self.followerList, sortFollowers)
    for _, followerData in pairs( self.followerList ) do
      if followerData.isCollected and ( followerData.status ~= "Inactive" or self.db.profile.showInactivesMatrix ) then
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        local counters = {}
        for id, abilityData in pairs( abilities ) do

          local counterID, _ = next(abilityData.counters)
          if not abilityData.isTrait then
            table.insert( counters, counterID)
            counterSum[counterID] = ( counterSum[counterID] or 0 ) + 1
            tooltip:SetCell(2, tooltip.Columns[tostring(counterID)].col, counterSum[counterID])

          end
        end
        if #counters > 1 then
          addDualCounter( counters[1], counters[2], followerData.iLevel )
        end
      end
    end
    local r,c
    for r, cData in pairs(tooltip.counterDualSum) do
      for c, rData in pairs(cData) do
        local formatedNumber
        if self.db.profile.colorMatrix then
          formatedNumber = string.format(self.formatStrings.rarityColors[rData.quality], rData.num)
        else
          formatedNumber = rData.num
        end
        tooltip:SetCell( tooltip.Rows[r], tooltip.Columns[tostring(c)].col, formatedNumber)


      end
    end
  else
    self.matrixIsUsed = false
    -- Tooltip: List of followers with 5 columns for counters/traits
    tooltip:AddHeader()
    table.sort( self.followerList, sortFollowers)

    local reRollCombos = {}
    if self.db.profile.showRerollPercent then
      local rerollIcon =  string.format(self.formatStrings.counterIconTolltip, GetItemIcon(118354))
      MyAddColumn("reroll")
      tooltip:SetCell(1, tooltip.Columns["reroll"].col, rerollIcon)
      for _, followerData in pairs( self.followerList ) do
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        local counters = {}
        for id, abilityData in pairs( abilities ) do
          local counterID, _ = next(abilityData.counters)
          if not abilityData.isTrait then
            table.insert( counters, counterID)
          end
        end
        if #counters == 2 then
          table.sort(counters)
          local combo = counters[1]+counters[2]/100
          followerData.combo  = combo
          reRollCombos[combo] = ( reRollCombos[combo] or 0 ) +1
        end
      end
    end

    MyAddColumn("counter1","")
    MyAddColumn("counter2","")
    AddSpanner("Threats", "vs", tooltip.Columns["counter1"].col, tooltip.Columns["counter2"].col)
    MyAddColumn("trait1","")
    MyAddColumn("trait2","")
    MyAddColumn("trait3","")
    AddSpanner("Traits", "Traits", tooltip.Columns["trait1"].col, tooltip.Columns["trait3"].col)

    for _, followerData in pairs(  self.followerList ) do
      if followerData.isCollected then
        local row = tooltip:AddLine("")
        local coloredName = string.format(self.formatStrings.rarityColors[followerData.quality], followerData.name)
        tooltip:SetCell(row, 1, string.format("%s (%d)%s", coloredName, followerData.level == 100 and followerData.iLevel or followerData.level, (followerData.status == "Inactive" and " X" or "") ) )
        if self.db.profile.showRerollPercent and followerData.combo and reRollCombos[followerData.combo] > 1 then
          local reroll = self:GetRerollPercent( followerData, reRollCombos )
          tooltip:SetCell(row, tooltip.Columns["reroll"].col, reroll)
        end
        local counter, trait = 1,1
        local abilities = C_Garrison.GetFollowerAbilities(followerData.followerID)
        table.sort(abilities, sortAbilities)

        for id, abilityData in pairs( abilities ) do
          local _, counterData = next(abilityData.counters)

          if  abilityData.isTrait then
            tooltip:SetCell(row, tooltip.Columns["trait"..trait].col, string.format(self.formatStrings.counterIconTolltip, counterData and counterData.icon or abilityData.icon))
            trait = trait+1
          else
            tooltip:SetCell(row, tooltip.Columns["counter"..counter].col, string.format(self.formatStrings.counterIconTolltip, counterData and counterData.icon or abilityData.icon) )
            counter = counter+1
          end
        end
      end
    end
  end
end

function BF:GetRerollPercent( followerData, reRollCombos )
  local spec = self.specs[followerData.classSpec] or self.specs[followerData.className]
  local allCombos, wantedCombos = 0, 0
  for i=1,(#spec-1) do
    for j=i+1,#spec do
      local combo = spec[i] + spec[j]/100
      allCombos = allCombos +1
      if not reRollCombos[combo] then
        wantedCombos = wantedCombos+1
      end
    end
  end
  local percentage = ( allCombos==0 ) and 0 or (  wantedCombos / allCombos * 100 )
  return math.floor(percentage).."%"
end

--
function BF:OnEnable()
  self:RegisterEvent("GARRISON_FOLLOWER_XP_CHANGED",    "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_ADDED",         "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_LIST_UPDATE",   "RedrawTooltip")
  self:RegisterEvent("GARRISON_FOLLOWER_REMOVED",       "RedrawTooltip")

  self:RegisterEvent("GARRISON_RECRUITMENT_NPC_OPENED", "EnhanceRecruitFrame")

  self:RegisterEvent("GARRISON_MISSION_NPC_OPENED",     "EnhanceFollowerTab")
end

function BF:OnDisable()
  self:UnregisterEvent("GARRISON_FOLLOWER_XP_CHANGED"   )
  self:UnregisterEvent("GARRISON_FOLLOWER_ADDED"        )
  self:UnregisterEvent("GARRISON_FOLLOWER_LIST_UPDATE"  )
  self:UnregisterEvent("GARRISON_FOLLOWER_REMOVED"      )
  self:UnregisterEvent("GARRISON_RECRUITMENT_NPC_CLOSED")
  self:UnregisterEvent("GARRISON_MISSION_NPC_OPENED"    )
end

---- Parse Follower List

function BF:UpdateFollowerCounters( ... )
  local followerData, counterID, followerAbilitiesData
  self.followerList = C_Garrison.GetFollowers()
  wipe( self.collectedFollowersList )
  wipe( self.reverseCounterLookup )
  self.collectedFollowersList = {}
  self.reverseCounterLookup   = {}

  for _, followerData in pairs( self.followerList ) do
    counters = {}

    for _, followerAbilitiesData in ipairs(C_Garrison.GetFollowerAbilities(followerData.followerID)) do
      counterID, _ = next( followerAbilitiesData.counters)

      if counterID then
        counters[counterID] = true

        self.reverseCounterLookup[counterID] = self.reverseCounterLookup[counterID] or {}
        tinsert(self.reverseCounterLookup[counterID], followerData.followerID)
      end
    end
    self.collectedFollowersList[followerData.followerID] =  counters
  end
end

---- Enhancing the Inn Recruitment

function BF:EnhanceRecruitFrame()
  --if true then return end
  local recruitFrame = GarrisonRecruitSelectFrame.FollowerSelection
  local followers = C_Garrison.GetAvailableRecruits();
  for i=1, 3 do
    recruitFrame["Recruit"..i].PortraitFrame:SetScript("OnEnter",
      function(tt)
        if not BF.db.profile.availableCountersToolstip then return end
        local tooltip = LibQTip:Acquire("Broker_Followers_EnhancementTooltip", 1, "LEFT")
        BF.enhanceTooltip = tooltip
        BF:AdditionalAbilitesTooltip(tooltip, followers[i])
        --tooltip:SmartAnchorTo(tt)
        tooltip:SetPoint("RIGHT",tt,"LEFT")
        tooltip:Show()
      end
    )
    recruitFrame["Recruit"..i].PortraitFrame:SetScript("OnLeave",
      function()
        if not BF.db.profile.availableCountersToolstip then return end
        LibQTip:Release(BF.enhanceTooltip)
        BF.enhanceTooltip.Columns = nil
        BF.enhanceTooltip.Rows = nil
        BF.enhanceTooltip = nil
      end
    )
  end
end

function BF:EnhanceFollowerTab()
  local followerTabPortrait = GarrisonMissionFrame.FollowerTab.PortraitFrame
  followerTabPortrait:SetScript("OnEnter",
    function(tt)
      local tooltip = LibQTip:Acquire("Broker_Followers_EnhancementTooltip", 1, "LEFT")
      BF.enhanceTooltip = tooltip
      BF:AdditionalAbilitesTooltip(tooltip, C_Garrison.GetFollowerInfo(GarrisonMissionFrame.FollowerTab.followerID))
      tooltip:SetPoint("RIGHT",tt,"LEFT")
      tooltip:Show()
    end)
  followerTabPortrait:SetScript("OnLeave",
    function()
    LibQTip:Release(BF.enhanceTooltip)
    BF.enhanceTooltip = nil
  end )
end

function BF:AdditionalAbilitesTooltip( tooltip, followerData)
  local possibleAbilities = self.specs[followerData.classSpec] or self.specs[followerData.className]
  for _, aid in pairs(possibleAbilities) do
    local tex = select(3, getMechanicInfo( aid ))
    tex = string.format(self.formatStrings.counterIconTolltip, tex)
    tooltip:AddHeader(tex)
  end
end

---- Database Maintenance

function BF:MigrateSavedData()
  local  globaldb = self.db.global
  local profiledb = self.db.profile

  -- all db sections are current we can exit immediately
  if profiledb.version == DB_VERSION and
      globaldb.version == DB_VERSION then
    return
  else
    BF:Print(L["DB_VERSION_UPGRADE_PURGE"])
  end

  -- Migrating from before addon began keeping track of database changes
  --[[
  if globaldb.version < X then
    -- Upgrade from before version X
  end
  if profiledb.version < X then
    -- Upgrade from before version X
  end
  --]]--

  profiledb.version = DB_VERSION
  globaldb.version = DB_VERSION
end

function BF:Debug(...)
  DebugPrint( ... )
end